import asyncio
import json
import mimetypes
import os
import re
from asyncio import run

import tornado

from . import token
from .util import getLogger


class PublicJSONHandler(tornado.web.RequestHandler):
    def prepare(self):
        self._logger = None
        self._data = None
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "*")
        self.set_header("Access-Control-Allow-Methods", "*")

    def logger(self):
        if self._logger is None:
            self._logger = getLogger(f"handler:{self.request.uri}")
        return self._logger

    def set_default_headers(self):
        self.set_header("Content-Type", "application/json")

    def jwt(self):
        auth_header = self.request.headers.get("Authorization")
        if auth_header is not None:
            return re.sub("^Bearer +", "", auth_header)
        return None

    def param(self, param_name, default=None):
        if self._data is None:
            try:
                self._data = json.loads(self.request.body)
            except json.JSONDecodeError:
                self._data = {}
        return self._data.get(param_name, self.get_argument(param_name, default))

    def file(self, fpath, mimetype=None):
        assert os.path.exists(fpath), f"Path {fpath} does not exist"
        assert os.path.isfile(fpath), f"Path {fpath} is not a file"

        if mimetype is None:
            content_type, encoding = mimetypes.guess_type(fpath)
            mimetype = content_type

        assert mimetype, f"Could not infer mimetype of {fpath} and no default provided"

        self.set_header("Content-Type", mimetype)
        self.set_header("Content-Length", os.path.getsize(fpath))

        with open(fpath, "rb") as f:
            while True:
                chunk = f.read(65 * 1024)
                if not chunk:
                    break
                self.write(chunk)
                self.flush()

        self.finish()

    def jsonerr(self, message, status=500):
        self.json({"status": "error", "message": message}, status)
        self.finish()

    def json(self, data, status=None):
        if status is not None:
            self.set_status(status)
        return self.write(json.dumps(data))

    def options(self):
        return self.json({"status": "ok"})


class JSONHandler(PublicJSONHandler):
    def prepare(self):
        self._logger = None
        self._data = None
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "*")
        self.set_header("Access-Control-Allow-Methods", "*")
        if self.jwt() is None:
            self.json({"status": "error", "message": "forbidden"}, 403)
            self.finish()
            return
        try:
            token.decode(self.jwt())
        except Exception:
            self.json({"status": "error", "message": "forbidden"}, 403)
            self.finish()
            return

    def decodedJwt(self):
        return token.decode(self.jwt())


class Default404Handler(PublicJSONHandler):
    def prepare(self):
        self.json({"status": "error", "message": "not found"}, status=404)
        self.finish()
        return self.request.connection.close()


async def start(
    name,
    port,
    routes,
    static_path=None,
    static_url_prefix=None,
    default_handler_class=None,
    debug=False,
):
    if default_handler_class is None:
        default_handler_class = Default404Handler
    app = tornado.web.Application(
        routes,
        default_handler_class=default_handler_class,
        debug=debug,
        static_path=static_path,
        static_url_prefix=static_url_prefix,
    )
    app.logger = getLogger(name)
    app.logger.info(f"  listening on {port}...")
    app.listen(int(port))
    await asyncio.Event().wait()



import unittest

from src.pycronado import core


class TestBasics(unittest.TestCase):
    def test_basics(self):
        core.getLogger
        self.assertTrue(True)
        self.assertIsNone(None)
        self.assertEqual("a", "a")



import logging


def getLogger(name, level=logging.DEBUG):
    logger = logging.getLogger(name)
    logger.setLevel(level)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    return logger



from os import environ as ENV

import jwt
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from jwt.exceptions import InvalidTokenError

JWT_PRIVKEY = ENV.get("JWT_PRIVKEY")
JWT_PRIVKEY_PASSWORD = ENV.get("JWT_PRIVKEY_PASSWORD")
JWT_PUBKEY = ENV.get("JWT_PUBKEY")


def ensureKeysFromPEM(private_key_pem):
    global JWT_PRIVKEY
    global JWT_PUBKEY
    JWT_PRIVKEY = decodePrivateKey(private_key_pem)
    JWT_PUBKEY = mkPublicKey(private_key_pem)


def ensureFreshKeys(key_size=4096):
    ensureKeysFromPEM(mkPrivateKey())


def decodePrivateKey(private_key_pem):
    return serialization.load_pem_private_key(
        private_key_pem, password=JWT_PRIVKEY_PASSWORD
    )


def mkPrivateKey(key_size=4096):
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size)

    if JWT_PRIVKEY_PASSWORD:
        encryption_algorithm = serialization.BestAvailableEncryption(
            JWT_PRIVKEY_PASSWORD
        )
    else:
        encryption_algorithm = serialization.NoEncryption()

    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=encryption_algorithm,
    )

    return pem


def mkPublicKey(private_key_pem):
    private_key = decodePrivateKey(private_key_pem)
    public_key = private_key.public_key()
    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    return public_key_pem


def encode(object):
    assert JWT_PRIVKEY, "No private key in JWT_PRIVKEY"
    return jwt.encode(object, JWT_PRIVKEY, algorithm="RS256")


def decode(tok):
    assert JWT_PUBKEY, "Need public key in JWT_PUBKEY"
    return jwt.decode(
        tok,
        JWT_PUBKEY,
        algorithms=["RS256"],
    )


def hasValidSig(token):
    """
    Checks for a valid signature on the given token, disregards expiration status.
    This is useful for checking incoming JWTs for refreshing, use `decode` for
    """
    try:
        jwt.decode(
            token,
            JWT_PUBKEY,
            algorithms=["RS256"],
            options={"verify_exp": False, "verify_iat": False},
        )
        return True

    except InvalidTokenError:
        return False
